;*=====================================================================*/
;*    .../project/bigloo/api/multimedia/examples/musicplay.scm         */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Thu Jan 31 11:44:28 2008                          */
;*    Last change :  Mon Jun  2 11:05:20 2008 (serrano)                */
;*    Copyright   :  2008 Manuel Serrano                               */
;*    -------------------------------------------------------------    */
;*    A simple music player                                            */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module musicplay
   (library multimedia)
   (main main))

;*---------------------------------------------------------------------*/
;*    main ...                                                         */
;*---------------------------------------------------------------------*/
(define (main args)
   (let ((files '())
	 (backend 'mplayer)
	 (command #f))
      (args-parse (cdr args)
	 (("--mpg123" (help "Select the mpg123 back-end"))
	  (set! backend 'mpg123))
	 (("--mpc" (help "Select the mpc back-end"))
	  (set! backend 'mpc))
	 (("--mplayer" (help "Select the mplayer back-end"))
	  (set! backend 'mplayer))
	 (("--command" ?cmd (help "Set the command path"))
	  (set! command cmd))
	 (("--help" (help "This help"))
	  (print "usage: music [options] file ...")
	  (args-parse-usage #f)
	  (exit 0))
	 (else
	  (set! files (cons else files))))
      (let ((player (case backend
		       ((mpg123)
			(if command
			    (instantiate::mpg123
			       (path command))
			    (instantiate::mpg123)))
		       ((mplayer)
			(if command
			    (instantiate::mplayer
			       (path command))
			    (instantiate::mplayer)))
		       ((mpc)
			(instantiate::mpc)))))
	 (for-each (lambda (p) (music-playlist-add! player p)) (reverse files))
	 (music-play player)
	 (music-event-loop player
            :onstate (lambda (status)
			(with-access::musicstatus status (state song volume)
			   (print "state   : " state)
			   (print "song    : " song)))
	    :onmeta (lambda (meta playlist)
		       (print "meta    : " meta)
		       (print "playlist: ")
		       (for-each (lambda (s) (print "        " s)) playlist))
	    :onvolume (lambda (volume)
		       (print "volume  : " volume))))))

